#pragma once

#include "IFontEngine.h"

#include "Image.h"

#include <iostream>
#include <filesystem>

#include <ft2build.h>
#include FT_FREETYPE_H

class FreeTypeFontEngine : public IFontEngine
{
public:
    FreeTypeFontEngine() = default;

    void initialize()
    {
        if (mLibrary)
        {
            return;
        }

        const auto error = FT_Init_FreeType(&mLibrary);
        if (error)
        {
            std::cout << "Error initializing FreeType" << std::endl;
        }
    }

    void loadFontFace(const std::filesystem::path& fontFile)
    {
        const auto full_path = mSystemFontLocation / fontFile;
        const auto error = FT_New_Face(mLibrary, full_path.c_str(), 0, &mWorkingFontFace );
        if ( error == FT_Err_Unknown_File_Format )
        {
            std::cout << "Found file but format not supported" << std::endl;
        }
        else if ( error )
        {
            std::cout << "Failed to find or open file" << std::endl;
        }

        int pensize = 16;
        FT_Set_Pixel_Sizes(mWorkingFontFace, pensize, 0);
    }

    std::unique_ptr<Image<unsigned char> > loadGlyph(unsigned charCode)
    {
        auto glyph_index = FT_Get_Char_Index( mWorkingFontFace, charCode );
        if (glyph_index == 0)
        {
            std::cout << "Got the null glyph" << std::endl;
            return nullptr;
        }

        auto error = FT_Load_Glyph(mWorkingFontFace, glyph_index, FT_LOAD_DEFAULT );
        if (error == FT_Err_Invalid_Argument)
        {
            std::cout << "Error loading glyph - bad arg" << std::endl;
            return nullptr;
        }
        else if(error)
        {
            auto code = static_cast<int>(error);
            std::cout << "Error loading glyph - somethign else " << code <<  std::endl;
            return nullptr;
        }

        if (mWorkingFontFace->glyph->format == FT_GLYPH_FORMAT_BITMAP)
        {
            std::cout << "It is bitmap format" << std::endl;
        }
        else
        {
            error = FT_Render_Glyph( mWorkingFontFace->glyph, FT_RENDER_MODE_NORMAL );
        }

        auto rows = mWorkingFontFace->glyph->bitmap.rows;
        auto columns = mWorkingFontFace->glyph->bitmap.width;
        std::cout << "We have a bitmap with rows " << rows << " and columns " << columns << std::endl;

        auto image = std::make_unique<Image<unsigned char>> (columns, rows);

        unsigned counter = 0;
        std::vector<unsigned char> data(rows*columns);
        for (unsigned idx=0; idx<rows; idx++)
        {
            for (unsigned jdx=0; jdx<columns; jdx++)
            {
                data[counter] = mWorkingFontFace->glyph->bitmap.buffer[counter];
            }
        }
        image->setData(data);
        return image;
    }

private:
    std::filesystem::path mSystemFontLocation{"/usr/share/fonts/"};

    FT_Face mWorkingFontFace{nullptr};
    FT_Library mLibrary{nullptr};
};
